# -*- coding: utf-8 -*-

from aiohttp import ClientSession, TCPConnector
import asyncio
from cachetools import TTLCache
from requests import Session

from .api_toolkit import make_headers
from .async_init import AsyncInitObject
from .cache_utils import self_cache

from typing import (
    Any,
    Coroutine,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

__all__ = (
    "AsyncSession",
    "SyncSession",
)


# XXX: in both functions I need to find a suitable cache_limit
# 1024 is a relatively random choice and
# has nothing to do with the desired behavior

class AsyncSession(AsyncInitObject):
    async def __init__(self, token: str, trust_env: bool = True,
                       cache_ttl: Union[int, float] = 60,
                       cache_limit: int = 1024,
                       use_cache: bool = True) -> None:
        headers = make_headers(token)
        loop = asyncio.get_event_loop()
        self.session = ClientSession(
            loop=loop,
            connector=TCPConnector(use_dns_cache=False, loop=loop),
            trust_env=trust_env,
            headers=headers,
        )

        self.cache = TTLCache(maxsize=cache_limit, ttl=cache_ttl)
        self.use_cache = use_cache

    async def close(self) -> None:
        """Close underlying connector.
        Release all acquired resources.
        """
        if not self.closed:
            # XXX: https://github.com/aio-libs/aiohttp/issues/1925
            await self.session.close()
            # Zero-sleep to allow underlying connections to close
            await asyncio.sleep(0)

    @property
    def closed(self) -> bool:
        """Is client session closed.
        A readonly property.
        """
        return self.session.closed

    async def simple_get_json(self, url: str) -> Dict:
        data = {}
        async with self.session.get(url) as response:
            data = await response.json()
        return data

    @self_cache(sync=False)
    async def cached_get_json(self, url: str) -> Dict:
        return await self.simple_get_json(url)

    async def get_json(self, url: str,
                       use_cache: Optional[bool] = None) -> Dict:
        if use_cache is None:
            use_cache = self.use_cache

        if use_cache:
            return await self.cached_get_json(url)
        else:
            return await self.simple_get_json(url)


class SyncSession:
    def __init__(self, token: str, trust_env: bool = True,
                 cache_ttl: Union[int, float] = 60,
                 cache_limit: int = 1024, use_cache: bool = True) -> None:
        self._closed = False

        headers = make_headers(token)
        self.session = Session()
        self.session.trust_env = trust_env
        self.session.headers.update(headers)

        self.cache = TTLCache(maxsize=cache_limit, ttl=cache_ttl)
        self.use_cache = use_cache

    def close(self) -> None:
        """Closes all adapters and as such the session"""
        if not self.closed:
            self.session.close()
            self._closed = True

    @property
    def closed(self) -> bool:
        """Is client session closed.
        A readonly property.
        """
        return self._closed

    def simple_get_json(self, url: str) -> Dict:
        with self.session.get(url) as response:
            data = response.json()

        return data

    @self_cache(sync=True)
    def cached_get_json(self, url: str) -> Dict:
        return self.simple_get_json(url)

    def get_json(self, url: str, use_cache: Optional[bool] = None) -> Dict:
        if use_cache is None:
            use_cache = self.use_cache

        if use_cache:
            return self.cached_get_json(url)
        else:
            return self.simple_get_json(url)



from cachetools import keys, Cache
from functools import wraps

__all__ = (
    "async_cached",
    "self_cache"
)


# annotating the remaining parameters of this function causes too many problems
def async_cached(cache: Cache, key=keys.hashkey, lock=None):
    def decorator(func):
        if lock is None:
            @wraps(func)
            async def wrapper(*args, **kwargs):
                k = key(*args, **kwargs)
                try:
                    return cache[k]
                except KeyError:
                    pass  # key not found

                val = await func(*args, **kwargs)

                try:
                    cache[k] = val
                except ValueError:
                    pass  # val too large

                return val
        else:
            @wraps(func)
            async def wrapper(*args, **kwargs):
                k = key(*args, **kwargs)
                try:
                    async with lock:
                        return cache[k]
                except KeyError:
                    pass  # key not found

                val = await func(*args, **kwargs)

                try:
                    async with lock:
                        cache[k] = val
                except ValueError:
                    pass  # val too large

                return val
        return wrapper
    return decorator


def self_cache(*, sync: bool):
    def decorator(func):
        if not sync:
            @wraps(func)
            async def wrapper(self, *args, **kwargs):
                new_func = async_cached(cache=self.cache)(func)
                return await new_func(self, *args, **kwargs)
        else:
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                new_func = cached(cache=self.cache)(func)
                return new_func(self, *args, **kwargs)

        return wrapper
    return decorator



# -*- coding: utf-8 -*-

import asyncio

from .async_init import AsyncInitObject
from .sessions import AsyncSession, SyncSession

from types import TracebackType
from typing import (
    Any,
    Coroutine,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

__all__ = (
    "AsyncClient",
    "SyncClient",
    "make_headers",
)


class AsyncClient(AsyncInitObject):
    async def __init__(self, token: str) -> None:
        self.session = await AsyncSession(token)

    async def close(self) -> None:
        """Close session"""
        await self.session.close()

    @property
    def closed(self) -> bool:
        """Is client session closed.
        A readonly property.
        """
        return self.session.closed

    def __enter__(self) -> None:
        raise TypeError("Use `async with` instead")

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        # __exit__ should exist in pair with __enter__ but never executed
        pass

    async def __aenter__(self) -> "AsyncClient":
        return self

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> None:
        await self.close()


class SyncClient:
    def __init__(self, token: str) -> None:
        self.session = SyncSession(token)

    def close(self) -> None:
        """Close session"""
        self.session.close()

    @property
    def closed(self) -> bool:
        """Is client session closed.
        A readonly property.
        """
        return self.session.closed

    def __enter__(self) -> "SyncClient":
        return self

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        self.close()

    async def __aenter__(self) -> None:
        raise TypeError("Use `with` instead")

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> None:
        # __aexit__ should exist in pair with __aenter__ but never executed
        pass



import sys

from . import __version__, __name__

from typing import Dict, Union


def make_headers(token: str) -> Dict[str, str]:
    return {
        "dnt": "1",
        "authorization": f"Bearer {token}",
        "user-agent": f"{__name__}/{__version__} (Python {sys.version[:5]})",
        "accept-encoding": ", ".join(("br", "gzip", "deflate")),
        "cache-control": "no-cache",
        "pragma": "no-cache",
    }



class ClientException(Exception):
    """Base class for all client exceptions."""


class ClientResponseError(ClientException):
    """Connection error during reading response."""

    def __init__(self, url: str, reason: str = "", message: str = ""):
        self.url = url
        self.reason = reason
        self.message = message


class UnexpectedResponseCode(ClientResponseError):
    """Occurs if the response code was not described
    in the official api documentation.
    """

    def __init__(self, code: int, **kwargs):
        self.code = code
        super().__init__(**kwargs)


class BadRequest(ClientResponseError):
    """Client provided incorrect parameters for the request."""

    code = 400


class Forbidden(ClientResponseError):
    """Access denied, either because of missing/incorrect credentials or
    used API token does not grant access to the requested resource.
    """

    code = 403


class NotFound(ClientResponseError):
    """Resource was not found."""

    code = 404


class TooManyRequests(ClientResponseError):
    """Request was throttled, because amount of requests
    was above the threshold defined for the used API token.
    """

    code = 429


class InternalServerError(ClientResponseError):
    """Unknown error happened when handling the request."""

    code = 500


class ServiceUnavailable(ClientResponseError):
    """Service is temprorarily unavailable because of maintenance."""

    code = 503
